/* Copyright (c) 2002-2005, The Ohio State University. All rights
 * reserved.
 *
 * This file is part of the MVAPICH software package developed by the
 * team members of The Ohio State University's Network-Based Computing
 * Laboratory (NBCL), headed by Professor Dhabaleswar K. (DK) Panda.
 *
 * For detailed copyright and licencing information, please refer to the
 * copyright file MVAPICH_COPYRIGHT in the top level MPICH directory.
 *
 */

#ifndef _COLLUTILS_H_
#define _COLLUTILS_H_

#ifndef VIADEV_RPUT_SUPPORT
#define VIADEV_RPUT_SUPPORT
#endif

#ifdef VIADEV_RPUT_SUPPORT

#include "vapi_header.h"
#include "vbuf.h"
#include "dreg.h"
#include "mpid_bind.h"
#include "comm.h"

/* Utility macros */

#define CHECK_ALLOC(ptr) {                                          \
    if(NULL == ptr) {                                               \
        fprintf(stderr, "No memory at line %d in file %s\n",        \
                __LINE__, __FILE__);                                \
        exit(-1);                                                   \
    }                                                               \
}

/* This flag is to indicate the arrival of
 * a RDMA message */
#define ALLTOALL_RDMA_FLAG (111)

/* Space reserved in the collective buffer for keeping flags */
#define COLL_FLAG_SPACE     (32)

/* Maximum number of DE processes */
#define MAX_DE_NODES        (viadev.np)

/* Space for the address & remote keys to be sent */
#define COLL_ADDR_SEND_SPACE ((sizeof(aint_t)+sizeof(VAPI_rkey_t) \
        + sizeof(char)))

#define COLL_ACK_SPACE      (2*MAX_DE_NODES*(sizeof(char)))

/* The Direct Eager address buffer should hold the destination
 * addresses of the receive buffers for total N processes
 * participating in the Alltoall.
 */
#define COLL_ADDR_RECV_SPACE ((2 * MAX_DE_NODES * (sizeof(aint_t) +  \
            sizeof(VAPI_rkey_t) + sizeof(char))))


#define BARRIER_SPACE   2048

#define COLL_BUF_SIZE       (COLL_FLAG_SPACE + COLL_ADDR_SEND_SPACE \
        + COLL_ACK_SPACE + COLL_ADDR_RECV_SPACE + 1024 +BARRIER_SPACE)

#define BARRIER_OFFSET (COLL_FLAG_SPACE + COLL_ADDR_SEND_SPACE \
        + COLL_ACK_SPACE + COLL_ADDR_RECV_SPACE + 1024)
#define ADDR_EXCHANGE_TAG   (101)

/* Collective structures */
struct R_Collbuf {
    void *buf;
    VIP_MEM_HANDLE buf_hndl;
};

struct L_Collbuf {
    void *buf;
    VIP_MEM_HANDLE buf_hndl;
};

struct DE_Remote_Node {
    /* Here is where this node's address
     * will be received */
    void *recv_addr[2];

    /* Here is where I will write to the remote
     * node's memory */
    void *dest_addr[2];

    /* This tells me if this remote node is 
     * done with me or not */
    void *done[2];

    /* This is where I write to let the remote
     * node know that I am done with it */
    void *r_done[2];

    VIP_MEM_HANDLE memhandle;
};

struct Direct_Eager {
    int alternate;
    struct DE_Remote_Node *node;
};

struct Alltoall {
    void *rdma_flag;
    int rdma_flag_size;
    struct Direct_Eager *de;

    void *rdma_recv_addr;
    int rdma_recv_addr_len;

    void *buf_hndl_addr;
    int buf_hndl_addr_len;
};


struct Collbuf {
    struct R_Collbuf *r_coll;
    struct L_Collbuf *l_coll;

    /* for alltoall support */
    struct Alltoall *atoa;

    /*for rdma barrier */
    int rdma_barrier_id;
    int togle;
};

struct Coll_Addr_Exch {
    void *buf;
    VIP_MEM_HANDLE buf_hndl;
};

/* Prototypes for collective utilities */

int log2(int);
int power_two(int);
void register_mem(void *, int, VIP_MEM_HANDLE *);
void deregister_mem(VAPI_mr_hndl_t *);
void copy_recv_addr(void *, struct Collbuf *, VIP_MEM_HANDLE *);
void coll_rdma_write(int, void *, int, VIP_MEM_HANDLE *, void *,
                     VIP_MEM_HANDLE *);
void Next_Power_of_Degree(int, int, int *);
int MPID_rdma_write(void *buf,
                    int len,
                    int src_lrank,
                    int dest_grank,
                    int togle,
                    void *remote_address, VIP_MEM_HANDLE * r_handle,
                    VIP_MEM_HANDLE * l_handle);
int intra_RDMA_barrier(struct MPIR_COMMUNICATOR *comm);
void coll_all_send_complete(int dest);

#endif                          /* VIADEV_RPUT_SUPPORT */

#endif                          /* _COLLUTILS_H_ */
