/* Copyright (c) 2002-2005, The Ohio State University. All rights
 * reserved.
 *
 * This file is part of the MVAPICH software package developed by the
 * team members of The Ohio State University's Network-Based Computing
 * Laboratory (NBCL), headed by Professor Dhabaleswar K. (DK) Panda.
 *
 * For detailed copyright and licencing information, please refer to the
 * copyright file COPYRIGHT_MVAPICH in the top level MPICH directory.
 *
 */

/*
 * vapi_const.h 
 * 
 * This file contains "tunable parameters" for VAPI. A lot of work 
 * needs to be done to find out what are the right values, and in 
 * fact whether this is the right set of knobs. Many are settable 
 * from the command line and
 * passed to the individual processes as environment variables.
 */

#ifndef _VAPI_CONST_H
#define _VAPI_CONST_H

#ifdef CO_PRINT
#define co_print(arg...)  \
do {                        \
   printf("[%d]", viadev.me);  \
   printf(arg); \
} while (0)
#else
#define co_print(arg...)
#endif


#define MAX_RDMA_SIZE			(1048576)

/* maximum number of processes supported. */
#define MAX_PROCESS_NUM			(512) 

/* DEFAULT_ADDR_LEN is the size of an element in the address string.
 * It indeed indicates the number of digits in hcalid/qp_num. 
 * Keeping this at 8 , though normally max seen is qp number 
 * of length 6.
 */

/* The length of QP representation in characters, could be 4 HEX */
#define QPLEN_XDR               (8)

#define DEFAULT_ADDR_LEN		(8)

/* MAX_ADDR_LEN is the total length of numbers of Queue Pairs 
 * a process has plus its hcalid number. In a fully-connected
 * model, each process at most has (MAX_PROCESS_NUM-1) Queue
 * Pairs. Thus, the total length of Queue Pairs is
 * (MAX_PROCESS_NUM-1)*DEFAULT_ADDR_LEN. Plus hcalid,
 * we round this length up to MAX_PROCESS_NUM * DEFAULT_ADDR_LEN.
 * Changed on June 8, 2003
 */ 

#define MAX_ADDR_LEN		(MAX_PROCESS_NUM * DEFAULT_ADDR_LEN)

#define VIADEV_DEFAULT_MTU             (MTU1024)
#define VIADEV_MAX_RDMA_SIZE        (1048576)
#define VIADEV_DEFAULT_PORT            (1)
#define VIADEV_DEFAULT_QP_OUS_RD_ATOM  (1)

#define VIADEV_DEFAULT_PSN             (0)
#define VIADEV_DEFAULT_PKEY_IX         (0)
#define VIADEV_DEFAULT_P_KEY           (0x0)
#define VIADEV_DEFAULT_MIN_RNR_TIMER   (5)
#define VIADEV_DEFAULT_SERVICE_LEVEL   (0)
#define VIADEV_DEFAULT_TIME_OUT        (10)

#define VIADEV_DEFAULT_STATIC_RATE     (0)

#define VIADEV_DEFAULT_SRC_PATH_BITS   (0)

#define VIADEV_MAX_FAST_EAGER_SIZE 255

/* To handle congestion in larger clusters */
#define VIADEV_DEFAULT_RETRY_COUNT     (7)

/* Set to infinite */
#define VIADEV_DEFAULT_RNR_RETRY       (7)

/*
#define VIADEV_DEFAULT_R_KEY           (0x0)
#define VIADEV_DEFAULT_L_KEY           (0x0)
#define VIADEV_MAX_NUM_DHANDLE      (10000)
*/
#define VIADEV_DEFAULT_MAX_SG_LIST  (20)
#define  VIADEV_DEFAULT_MAX_RDMA_DST_OPS (16)

#define DEFAULT_PORT            (1)   
#define DEFAULT_QP_OUS_RD_ATOM  (1)   
#define DEFAULT_MTU             (MTU1024)
#define DEFAULT_PSN             (0)
#define DEFAULT_PKEY_IX         (0)
#define DEFAULT_P_KEY           (0x0)
#define DEFAULT_MIN_RNR_TIMER   (0x18)
#define DEFAULT_SERVICE_LEVEL   (0)
#define DEFAULT_TIME_OUT        (10)  

#define DEFAULT_STATIC_RATE     (0)    

#define DEFAULT_SRC_PATH_BITS   (0)

/* To handle congestion in larger clusters */
#define DEFAULT_RETRY_COUNT     (7)  

/* Set to infinite */
#define DEFAULT_RNR_RETRY       (7)

#if 0
#define DEFAULT_R_KEY           (0x0)
#define DEFAULT_L_KEY           (0x0)
#define MAX_NUM_DHANDLE         (10000)
#endif

#define DEFAULT_MAX_SG_LIST		(20)

#define DEFAULT_MAX_RDMA_DST_OPS (16)

#if 0
#define NUM_ADDR_BITS           (32);
#define MASK_ADDR_BITS          (0xffffffff);
#endif

#define HOSTNAME_LEN 			(255)

#ifdef MAC_OSX
#define VIADEV_NUM_RDMA_BUFFER 128
#else
#define VIADEV_NUM_RDMA_BUFFER 32 
#endif


#define VIADEV_PREPOST_DEPTH 		64	
#define VIADEV_VBUF_MIN_POOL_SIZE 	100
#define VIADEV_PREPOST_RENDEZVOUS_EXTRA VIADEV_PREPOST_DEPTH

#ifdef MCST_SUPPORT
#define DEFAULT_MAX_CQ_SIZE 	(40000)
#define DEFAULT_MAX_WQE    		(10000)
#define DEFAULT_FLOW_LABEL      (0)
#define DEFAULT_TRAFFIC_CLASS   (0)
#define DEFAULT_HOP_LIMIT       (63)
#endif


#endif /* _VAPI_CONST_H */
