/*
 * Copyright (C) 1999-2001 The Regents of the University of California
 * (through E.O. Lawrence Berkeley National Laboratory), subject to
 * approval by the U.S. Department of Energy.
 *
 * Use of this software is under license. The license agreement is included
 * in the file MVICH_LICENSE.TXT.
 *
 * Developed at Berkeley Lab as part of MVICH.
 *
 * Authors: Bill Saphir      <wcsaphir@lbl.gov>
 *          Michael Welcome  <mlwelcome@lbl.gov>
 */

/* Copyright (c) 2002-2005, The Ohio State University. All rights
 * reserved.
 *
 * This file is part of the MVAPICH software package developed by the
 * team members of The Ohio State University's Network-Based Computing
 * Laboratory (NBCL), headed by Professor Dhabaleswar K. (DK) Panda.
 *
 * For detailed copyright and licencing information, please refer to the
 * copyright file COPYRIGHT_MVAPICH in the top level MPICH directory.
 *
 */
/*
 * viaparam.h
 *
 * This file contains "tunable parameters" for MVICH. A lot of work needs to be
 * done to find out what are the right values, and in fact whether this is
 * the right set of knobs. Many are settable from the command line and
 * passed to the individual processes as environment variables.
 * Default value are set in the viaparam.c
 */



#ifndef _VIAPARAM_H
#define _VIAPARAM_H

#include "vapi_header.h"

#include <stdio.h>

#define MIN(a,b) ((a)<(b)?(a):(b))
#define MAX(a,b) ((a)>(b)?(a):(b))


/*
 * Currently, only Servernet requires us to limit the number
 * of outstanding RDMA operations in flight on any connection
 */
#define VIADEV_RDMA_LIMIT 2

#ifdef VIADEV_HAVE_RDMA_LIMIT
extern int viadev_rdma_limit;
#endif

/* the following are a list of globally defined variables to control
 * the behaviour and limits within MVICH.  Many are run-time setable.
 * See viaparam.c for details.
 */
extern char	*viadev_device[MAX_NUM_HCAS];


extern IB_mtu_t         viadev_default_mtu;
extern unsigned long        viadev_max_rdma_size;
extern int          viadev_default_port;
extern u_int8_t         viadev_default_qp_ous_rd_atom;
extern VAPI_psn_t       viadev_default_psn;
extern VAPI_pkey_ix_t       viadev_default_pkey_ix;
extern u_int32_t        viadev_default_p_key;
extern IB_rnr_nak_timer_code_t  viadev_default_min_rnr_timer;
extern IB_sl_t          viadev_default_service_level;
extern VAPI_timeout_t       viadev_default_time_out;
extern IB_static_rate_t     viadev_default_static_rate;
extern u_int8_t         viadev_default_src_path_bits;
extern int          viadev_max_fast_eager_size;
extern VAPI_retry_count_t   viadev_default_retry_count;
extern VAPI_retry_count_t   viadev_default_rnr_retry;
/*
extern VAPI_rkey_t      viadev_default_r_key;
extern VAPI_lkey_t      viadev_default_l_key;
extern unsigned int     viadev_max_num_dhandle;
*/
extern u_int32_t        viadev_default_max_sg_list;
extern u_int8_t         viadev_default_max_rdma_dst_ops;
extern unsigned int         viadev_ndreg_entries;

extern unsigned long viadev_sq_size;
extern unsigned long viadev_rq_size;

extern unsigned long viadev_cq_size;
extern int viadev_prepost_noop_extra;

extern int viadev_num_rdma_buffer;
extern int viadev_prepost_depth;
extern int viadev_prepost_threshold;
extern int viadev_initial_prepost_depth;
extern int viadev_initial_credits;
extern int viadev_vbuf_pool_size;
extern int viadev_vbuf_secondary_pool_size;
extern int viadev_vbuf_max;
extern int viadev_spin_count;
extern int viadev_rendezvous_threshold;
extern int viadev_prepost_rendezvous_extra;
extern int viadev_credit_preserve;
extern int viadev_credit_notify_threshold;
extern int viadev_dynamic_credit_threshold;

#ifdef VIADEV_PEER_TO_PEER
extern int viadev_use_peer_to_peer;
#endif


#ifdef _SMP_
extern int  smp_eagersize; /* EAGER threshold in SMP path */
extern int  smpi_length_queue; /* Shared memory buffer pool size */
#endif

/*
 * Why specify clock resolution in a parameters file?
 * It would be best to measure clock resolution, but
 * you are likely to measure slightly different values
 * on different nodes, and this is a Bad Thing. On most
 * recent systems, with Linux, gettimeofday() resolution
 * (MPID_Wtime is based on gettimeofday) is just a few
 * microseconds. The 10 microsecond value here is safe. 
 */

#define VIADEV_WTICK (0.000010)


/* Set default run-time parameters and over-ride with values
 * defined by environment variables passed into this process
 */
void viadev_init_parameters(int num_proc, int me);
void dump_param_values(FILE *fd, int me);

#endif /* _VIAPARAM_H */

